"""
This is the module that contains the main transformation from ufl to loopy
(with pdelab as the hardcoded generation target)
"""

from __future__ import absolute_import
from ufl.algorithms import MultiFunction
# Spread the pymbolic import statements to where they are used.
from pymbolic.primitives import Variable, Subscript, Sum, Product
import loopy
import numpy
import ufl

from dune.perftool import Restriction

# Define the generators that are used here
from dune.perftool.generation import generator_factory
loopy_iname = generator_factory(item_tags=("loopy", "kernel", "iname"))
loopy_expr_instruction = generator_factory(item_tags=("loopy", "kernel", "instruction", "exprinstruction"), no_deco=True)
loopy_temporary_variable = generator_factory(item_tags=("loopy", "kernel", "temporary"), on_store=lambda n: loopy.TemporaryVariable(n, dtype=numpy.float64), no_deco=True)
loopy_c_instruction = generator_factory(item_tags=("loopy", "kernel", "instruction", "cinstruction"), no_deco=True)
loopy_valuearg = generator_factory(item_tags=("loopy", "kernel", "argument", "valuearg"), on_store=lambda n: loopy.ValueArg(n), no_deco=True)

@generator_factory(item_tags=("loopy", "kernel", "argument", "globalarg"))
def loopy_globalarg(name, shape=loopy.auto):
    if isinstance(shape, str):
        shape = (shape,)
    return loopy.GlobalArg(name, numpy.float64, shape)

@generator_factory(item_tags=("loopy", "kernel", "domain"))
def loopy_domain(iname, shape):
    loopy_valuearg(shape)
    return "{{ [{0}] : 0<={0}<{1} }}".format(iname, shape)

@loopy_iname
def dimension_iname(index):
    from dune.perftool.pdelab import name_index
    from dune.perftool.pdelab.geometry import name_dimension
    iname = name_index(index)
    dimname = name_dimension()
    loopy_domain(iname, dimname)
    return iname

@loopy_iname
def argument_iname(arg):
    # TODO extract the {iname}_n thing by a preamble
    from dune.perftool.ufl.modified_terminals import ModifiedArgumentNumber
    iname = "arg{}".format(chr(ord("i") + ModifiedArgumentNumber()(arg)))
    loopy_domain(iname, iname + "_n")
    return iname

@loopy_iname
def quadrature_iname():
    loopy_domain("q", "q_n")
    return "q"

from dune.perftool.ufl.modified_terminals import ModifiedTerminalTracker
from dune.perftool.pymbolic.uflmapper import UFL2PymbolicMapper

class UFL2LoopyVisitor(ModifiedTerminalTracker, UFL2PymbolicMapper):
    def __init__(self, embracing_mf):
        super(UFL2LoopyVisitor, self).__init__()
        self.call = embracing_mf.call

    def coefficient(self, o):
        # All trial functions should already be handled by the TrialFunctionExtractor class!
        assert o.count() != 0
        # TODO implement non-trialfunction coefficients
        raise NotImplementedError

    # TODO use multiple inheritance and have a geometry transformer in the pdelab subpackage
    def facet_area(self, o):
        from pymbolic.primitives import Variable
        from dune.perftool.pdelab.geometry import name_facetarea
        return Variable(name_facetarea())


# TODO This pattern here could be abstracted as some sort of "SupersedingMultiFunction":
# A superseding multifunction is always asked to handle the given expression first. Only
# if it explicitly delegates, the underlying mf is called. Whenever, the underlying mf
# goes into recursion, it asks back to the superseding function. Martin would pity me for
# the performance overhead, but it makes life easier.
class TrialFunctionExtractor(MultiFunction):
    call = MultiFunction.__call__

    def __call__(self, o):
        from dune.perftool.ufl.modified_terminals import ModifiedArgumentExtractor
        self.tf = ModifiedArgumentExtractor()(o, trialfunction=True)
        self.u2l = UFL2LoopyVisitor(self)

        return self.call(o)

    def expr(self, o):
        if o in self.tf:
            # This is a modified trial function!
            from dune.perftool.pdelab.argument import name_trialfunction
            name = name_trialfunction(o)
            loopy_globalarg(name)
            return Variable(name)
        else:
            return self.u2l(o)

class _Counter:
    counter = 0

def get_count():
    c = _Counter.counter
    _Counter.counter = c + 1
    return c

def transform_accumulation_term(term):
    # Get the accumulation expression and the modified arguments
    expr, args = term

    # We always have a quadrature loop
    quadrature_iname()

    # Get the pymbolic expression needed for this accumulation term.
    # This includes filling the cache with all sorts of necessary preambles!
    pymbolic_expr = TrialFunctionExtractor()(expr)

    # Now simplify the expression
    # TODO: Add a switch to disable/configure this.
#     from dune.perftool.pymbolic.simplify import simplify_pymbolic_expression
#     pymbolic_expr = simplify_pymbolic_expression(pymbolic_expr)

    # Define a temporary variable for this expression
    expr_tv_name = "expr_"+str(get_count()).zfill(4)
    expr_tv = loopy_temporary_variable(expr_tv_name)
    loopy_expr_instruction(loopy.ExpressionInstruction(assignee=Variable(expr_tv_name), expression=pymbolic_expr))

    # The data that is used to collect the arguments for the accumulate function
    accumargs = []
    argument_code = []

    # Generate the code for the modified arguments:
    for arg in args:
        from dune.perftool.pdelab.argument import name_argumentspace, name_argument
        accumargs.append(name_argumentspace(arg))
        accumargs.append(argument_iname(arg))
        name = name_argument(arg)
        argument_code.append(name)
        loopy_globalarg(name)

    from dune.perftool.pdelab.argument import name_residual
    residual = name_residual()

    from dune.perftool.generation import retrieve_cache_items
    inames = retrieve_cache_items("iname")

    from dune.perftool.pdelab.quadrature import name_factor
    loopy_c_instruction(loopy.CInstruction(inames,
                                           "{}.accumulate({}, {}*{}*{})".format(residual,
                                                                                ", ".join(accumargs),
                                                                                expr_tv_name,
                                                                                "*".join(argument_code),
                                                                                name_factor()
                                                                                )
                                           )
                        )
